unit fDef;

interface

uses Classes;

type

{ TArrayOfSingle }

	TArrayOfSingle = array of Single;
  TArrayOfInteger= array of Integer;

{ TStkDataRec }//0000,13100,13100X(nonEtron),13171X(nonBank),17100
  PStkDataRec = ^TStkDataRec;
	TStkDataRec = packed record
    Date: TDateTime;					//
    OP,HP,LP,CP,VOL: Single;	//L,}L,̰,̧C,^,q
  end;

{ IBaseDataFile }
	IBaseDataFile = interface(IUnknown)
   	function getCount: Integer;
 		function getFileName: string;
    function getHeader: Pointer;
    function getHeaderSize: Integer;
    function getRec(Index: Integer): Pointer;
    function getRecSize: Integer;
    procedure loadFromFile(const FileName: string);
    procedure saveAs(FileName: string);
    function seek(Index: Integer): Pointer;
	end;

{ TBaseDataFile }
	TBaseDataFile = class(TInterfacedObject, IBaseDataFile)
  protected
	 	M: TMemoryStream;
  protected
   	function getCount: Integer; virtual;
    function getHeader: Pointer; virtual;
    function getRec(Index: Integer): Pointer; virtual;
    function seek(Index: Integer): Pointer; virtual;
    procedure loadFromFile(const FileName: string); virtual;
    procedure saveAs(FileName: string); virtual;
  protected
 		function getFileName: string; virtual; abstract;
    function getHeaderSize: Integer; virtual;
    function getRecSize: Integer; virtual; abstract;
  public
  	constructor Create; virtual;
    destructor Destroy; override;
  end;

{ IDataFile }
	//function getData(Date: WORD): Pointer;->Date<MAX_DATADAY means Index from tail toward header, or Search data by Date
	IDataFile = interface(IBaseDataFile)
    function getData(Date: WORD): Pointer;
    function getStockName: string;
    function indexOf(Date: WORD): Integer; overload;              //Search by Date, Index from header to tail
    function indexOf(Date: WORD; L,H: Integer): Integer; overload;//Search by Date, Index from header to tail
    function getCP: TArrayOfSingle;
    function getUD: TArrayOfSingle;
    function getVOL: TArrayOfSingle;
    procedure save;
  end;

{ TDataFile }
	TDataFile = class(TBaseDataFile, IDataFile)
  protected
  	FStockName: string;
    function FindData(Date: WORD; L,H: Integer): Pointer; virtual;
		function RecStart: Integer; virtual;
  protected//implementation of interface function//
		function getFileName: string; override;
		function getRecSize: Integer; override;
    function getData(Date: WORD): Pointer; virtual;
    function getStockName: string; virtual;
    function indexOf(Date: WORD): Integer; overload;              //Search by Date, Index from header to tail
    function indexOf(Date: WORD; L,H: Integer): Integer; overload;//Search by Date, Index from header to tail
    procedure save; virtual;
    function seek(Index: Integer): Pointer; override;
    function getCP: TArrayOfSingle;
    function getUD: TArrayOfSingle;
    function getVOL: TArrayOfSingle;
  public
  	constructor Create(const StockName: string); reintroduce;
  end;

implementation

uses fUtils, SysUtils;

{ TBaseDataFile }

constructor TBaseDataFile.Create;
begin
	inherited;
  loadFromFile(getFileName);
end;

destructor TBaseDataFile.Destroy;
begin
	_free_(M);
  inherited;
end;

function TBaseDataFile.getCount: Integer;
begin
	if M = nil then Result := 0 else Result := (M.Size-getHeaderSize) div getRecSize;
end;

function TBaseDataFile.getHeader: Pointer;
begin
	if M <> nil then Result := M.Memory else Result := nil;
end;

function TBaseDataFile.getHeaderSize: Integer;
begin
	Result := 0;
end;

function TBaseDataFile.getRec(Index: Integer): Pointer;
begin
	if (Index>-1) and (Index<getCount) then Result := Seek(Index) else Result := nil;
end;

procedure TBaseDataFile.loadFromFile(const FileName: string);
begin
  _free_(M);
  if FileExists(FileName) then
  begin
		M := TMemoryStream.Create;
  	M.LoadFromFile(FileName);
    M.Position := 0;
  end;
end;

procedure TBaseDataFile.saveAs(FileName: string);
begin
	if M <> nil then M.SaveToFile(FileName);
end;

function TBaseDataFile.seek(Index: Integer): Pointer;
begin
	if M = nil then Result := nil else Result := Pointer(Integer(getHeader) + getHeaderSize + Index * getRecSize);
end;

{ TDataFile }

constructor TDataFile.Create(const StockName: string);
begin
  FStockName := Trim(StockName);
	inherited Create;
end;

function TDataFile.getFileName: string;
begin
	Result := '.\DATA\Ku\'+FStockName+'.DAT';
end;

function TDataFile.FindData(Date: WORD; L, H: Integer): Pointer;
var
	M: Integer;
  D: WORD;
begin
	Result := nil;
  if L <= H then
  begin
  	M := (L+H) shr 1;
    D := PWORD(getRec(M))^;
    if Date = D then Result := getRec(M)
    else if Date < D then Result := FindData(Date,L,M-1)
    else Result := FindData(Date,M+1,H);
  end;
end;

function TDataFile.getData(Date: WORD): Pointer;
begin
	if Date > 20000 then
  	Result := FindData(Date,0,getCount-1)//Search by Date
  else Result := getRec(getCount-Date-1);//Index Record from tail toward header
end;

function TDataFile.getRecSize: Integer;
begin
	Result := SizeOf(TStkDataRec);
end;

function TDataFile.RecStart: Integer;
begin
	Result := 0;
end;

function TDataFile.indexOf(Date: WORD): Integer;begin	Result := indexOf(Date,0,getCount-1);				end;
function TDataFile.getStockName: string; 				begin	Result := FStockName;												end;
procedure TDataFile.save; 											begin	if M <> nil then M.SaveToFile(getFileName);	end;

function TDataFile.Seek(Index: Integer): Pointer;
begin
	if (M = nil) or (getCount = 0) then Result := nil
  else begin
  	Index := (Index + RecStart) mod getCount;
    Result := Pointer(Integer(M.Memory) + getHeaderSize + Index * getRecSize);
  end;
end;

function TDataFile.indexOf(Date: WORD; L, H: Integer): Integer;
var
	M: Integer;
  D: WORD;
begin
	Result := -1;
  if L <= H then
  begin
  	M := (L + H) shr 1;
    D := PWORD(getRec(M))^;
    if Date = D then Result := M
    else if Date < D then Result := indexOf(Date,L,M-1)
    else Result := indexOf(Date,M+1,H);
  end;
end;

function TDataFile.getCP: TArrayOfSingle;
var
	I: Integer;
begin
	SetLength(Result, getCount);
  for I := 0 to getCount-1 do
  	Result[I] := PStkDataRec(getRec(I)).CP;
end;

function TDataFile.getVOL: TArrayOfSingle;
var
	I: Integer;
begin
	SetLength(Result, getCount);
  for I := 0 to getCount-1 do
  	Result[I] := PStkDataRec(getRec(I)).VOL;
end;

function TDataFile.getUD: TArrayOfSingle;
var
	I: Integer;
begin
	SetLength(Result, getCount);
  for I := 0 to getCount-1 do
  	if I = 0 then Result[I] := 0
    else Result[I] := PStkDataRec(getRec(I)).CP-PStkDataRec(getRec(I-1)).CP;
end;

end.
